/******************************************************************************
  (c) Copyright 2002-2008, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      
 *****************************************************************************/
package org.netkernel.scheduler;

import java.util.*;
import java.util.Map.Entry;
import java.util.concurrent.locks.ReentrantLock;

import org.netkernel.urii.IEndpoint;
import org.netkernel.util.MultiMap;

/**
 * Atomic lock request and acquire. unlock might return a released previously
 * unaquired lock request if one exists. 
 * 
 * @author tab
 */
public class LockedEndpointTable
{
	/** map of lock keys to list of unaquired lock request references */
	private final Map<IEndpoint, List<RequestState>> mKeyToReferences;
	/** lock to control access to internal data structure */
	private ReentrantLock mSyncLock = new ReentrantLock(false);	
	
	/**
	 * Creates a new instance of AsynchLock
	 * 
	 * @param aKeyCount
	 *            size to make lock map
	 *@param aContentionCount
	 *            expected lock contention count- so as to allocate a list long
	 *            enough to hold them without needing to expand
	 */
	public LockedEndpointTable()
	{	mKeyToReferences = new HashMap<IEndpoint, List<RequestState>>();
	}

	/**
	 * Attempt to gain an exclusive lock on the given key, takes a reference
	 * object to hang onto if the lock fails to be granted that is then returned
	 * on an unlock when the lock has been granted
	 * 
	 * @param aEndpoint
	 *            the key object that we want to gain an exclusive lock on
	 * @param aState
	 *            a reference object for the request to lock
	 * @return true if the lock has been granted, false if not
	 */
	public boolean lock(IEndpoint aEndpoint, RequestState aState)
	{	boolean aquiredLock;
		mSyncLock.lock();
		List<RequestState> states = mKeyToReferences.get(aEndpoint);
		if (states == null)
		{	aquiredLock = true;
			states = new ArrayList<RequestState>(4);
			states.add(aState);
			mKeyToReferences.put(aEndpoint, states);
		}
		else
		{	aquiredLock = false;
			states.add(aState);
		}
		mSyncLock.unlock();
		return aquiredLock;
	}

	/**
	 * releases a lock on the given key. If this method is called after a
	 * lock has been granted then a lock may be released. 
	 *  The returned lock request reference must be acted
	 * on by the client and not ignored. If the released state had not
	 * acquired the lock before release then no new state will aquire the
	 * lock.
	 * 
	 * @param aEndpoint
	 *            the key object that we are releasing the exclusive lock on
	 * @return a lock request reference object to a previously ungranted lock
	 *         request, null if there is none.
	 */
	public RequestState unlock(IEndpoint aEndpoint, RequestState aState)
	{
		RequestState unlocked = null;
		mSyncLock.lock();
		List<RequestState> states = mKeyToReferences.get(aEndpoint);
		if (states!=null)
		{	
			int i=states.indexOf(aState);
			if (i>=0)
			{	states.remove(i);
			}
			int size=states.size();
			if (size==0)
			{	mKeyToReferences.remove(aEndpoint);
			}
			else if (i==0)
			{	unlocked=states.get(0);
			}
		}
		mSyncLock.unlock();
		return unlocked;
	}

	public MultiMap dump()
	{	MultiMap result = new MultiMap(16, 2);
		mSyncLock.lock();
		for (Entry<IEndpoint, List<RequestState>> e : mKeyToReferences.entrySet())
		{	IEndpoint key = e.getKey();
			List<RequestState> v = e.getValue();
			for (Object ref : v)
			{	result.put(key, ref);
			}
		}
		mSyncLock.unlock();
		return result;
	}
	
	public List<RequestState> clear(RootState aRoot)
	{	List<RequestState> states = new ArrayList<RequestState>();
		mSyncLock.lock();
		
		for (Iterator<Entry<IEndpoint, List<RequestState>>> j=mKeyToReferences.entrySet().iterator(); j.hasNext(); )
		{	Entry<IEndpoint, List<RequestState>> e=j.next();
			IEndpoint key = e.getKey();
			List<RequestState> v = e.getValue();
			for (Iterator<RequestState> i=v.iterator(); i.hasNext(); )
			{
				RequestState rs=i.next();
				if (rs.getRoot()==aRoot)
				{	states.add(rs);
					i.remove();
				}
			}
			if (v.size()==0)
			{	j.remove();
			}
		}
		mSyncLock.unlock();
		return states;
	}
}